from production.box import ProductionBox
from typing import List, Dict, Any, Union, Optional

from ._BoxConfiguration import BoxConfiguration
from ._ProductionTree import ProductionTree
from ._TreeConfig import TreeConfig
from ._ViewConfiguration import ViewConfiguration
from .boxfactory import ProductionBoxFactory
from .viewfactory import ViewFactory


class TreeBuilder(object):
    def __init__(self, box_factory, view_factory):
        # type: (ProductionBoxFactory, ViewFactory) -> None
        self.box_factory = box_factory
        self.view_factory = view_factory

    def build_trees(self, tree_configs, branch_configs, view_configs):
        # type: (List[TreeConfig], List[TreeConfig], List[ViewConfiguration]) -> List[ProductionTree]  # noqa
        views = {}
        for view in view_configs:
            views[view.name] = self.view_factory.build(view)
        self.box_factory.set_views(views)

        branches = {}
        for branch in branch_configs:
            branches[branch.name] = branch

        trees = []
        for tree_config in tree_configs:
            trees.append(self.build_tree(tree_config, branches))

        return trees

    def build_tree(self, tree_config, branches=None, tree_name=''):
        # type: (TreeConfig, Dict[str, Any], Optional[str]) -> ProductionTree

        tree_name = tree_config.name if tree_name == '' else tree_name
        boxes = self.build_all_boxes(tree_config.boxes, branches, tree_name)
        name_box_dict = self.get_name_box_dict(boxes)
        self.set_son_in_every_box(boxes, tree_config.boxes, name_box_dict)
        leaf_boxes = self.get_leaf_boxes(tree_config.boxes, name_box_dict)

        if isinstance(boxes[0], ProductionTree):
            return ProductionTree(tree_config.name, boxes[0].root, leaf_boxes)
        else:
            return ProductionTree(tree_config.name, boxes[0], leaf_boxes)

    def build_all_boxes(self, boxes, branches, tree_name=''):
        # type: (List[BoxConfiguration], Dict[str, str], Optional[str]) -> List[ProductionBox]

        ret = []
        for box_config in boxes:
            if box_config.type == "Branch":
                box = self.build_tree(branches[box_config.extra_config["BranchName"]], branches, tree_name)
                box.name = box_config.name
            else:
                box = self.box_factory.build(box_config)

            box.set_tree_name(tree_name)
            ret.append(box)
        return ret

    @staticmethod
    def get_name_box_dict(boxes):
        name_box_dict = {}
        for box in boxes:
            name_box_dict[box.name] = box
        return name_box_dict

    @staticmethod
    def set_son_in_every_box(boxes, box_configs, name_box_dict):
        # type: (List[ProductionBox], List[BoxConfiguration], Dict[str, Union[ProductionBox, ProductionTree]]) -> None

        used_boxes = {}
        for box_config in box_configs:
            sons = []
            for son_name in box_config.sons:
                if son_name not in name_box_dict:
                    error_message = "Box referencing invalid son: BoxName: {} - SonName: {}"
                    raise Exception(error_message.format(box_config.name, son_name))

                son_box = name_box_dict[son_name]
                used_boxes[son_box] = son_box
                if isinstance(son_box, ProductionTree):
                    sons.append(son_box.root)
                else:
                    sons.append(son_box)

            if isinstance(name_box_dict[box_config.name], ProductionTree):
                production_tree = name_box_dict[box_config.name]  # type: ProductionTree
                for leaf in production_tree.leaves:
                    leaf.sons = sons
            else:
                name_box_dict[box_config.name].sons = sons

        if len(boxes) == len(used_boxes):
            raise Exception("Tree with cycle (No root node found)")

        if len(boxes) - len(used_boxes) > 1:
            remaining = []
            for box in boxes:
                if box not in used_boxes:
                    remaining.append(box)

            remaining_names = ""
            for box in remaining:
                remaining_names += box.name + ", "
            remaining_names = remaining_names[:-2]
            raise Exception("Tree without single root node. Remaining nodes: {}".format(remaining_names))

    @staticmethod
    def get_leaf_boxes(box_configs, name_box_dict):
        leaf_boxes = []
        for box_config in box_configs:
            if len(box_config.sons) == 0:
                leaf_boxes.append(name_box_dict[box_config.name])

        return leaf_boxes



from abc import ABCMeta, abstractmethod
from datetime import datetime


class Clock(object):
    __meta__ = ABCMeta

    @abstractmethod
    def now(self):
        # type: () -> datetime
        raise NotImplementedError()

    @abstractmethod
    def utc_now(self):
        # type: () -> datetime
        raise NotImplementedError()



from logging import Logger
from typing import Dict

from production.box import CustomPropertyValueFilterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class CustomPropertyValueFilterBoxBuilder(BoxBuilder):
    def __init__(
        self,           # type: CustomPropertyValueFilterBoxBuilder
        loggers,        # type: Dict[str, Logger]
    ):
        # type (...) -> None

        super(CustomPropertyValueFilterBoxBuilder, self).__init__(loggers)

    def build(
        self,       # type: CustomPropertyValueFilterBoxBuilder
        config,     # type: BoxConfiguration
    ):
        # type: (BoxConfiguration) -> CustomPropertyValueFilterBox

        return CustomPropertyValueFilterBox(
            name=config.name,
            sons=config.sons,
            parameter_name=self.get_extra(config, "PropertyName", None),
            allowed_values=self.get_extra(config, "AllowedValues", None),
            excluded_values=self.get_extra(config, "ExcludedValues", None),
            logger=self.get_logger(config),
        )



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import typing

from cryptography.hazmat.primitives import padding

def check_pkcs7_padding(data: bytes) -> bool: ...
def check_ansix923_padding(data: bytes) -> bool: ...

class PKCS7PaddingContext(padding.PaddingContext):
    def __init__(self, block_size: int) -> None: ...
    def update(self, data: bytes) -> bytes: ...
    def finalize(self) -> bytes: ...

class ObjectIdentifier:
    def __init__(self, val: str) -> None: ...
    @property
    def dotted_string(self) -> str: ...
    @property
    def _name(self) -> str: ...

T = typing.TypeVar("T")



from typing import Dict, Any

from nreports.builders.pre_account import PreAccountBySeatsBuilder
from nreports_api.shared import JSON


class ConferenceItemsBySeatsBuilder(PreAccountBySeatsBuilder):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        data_to_build = super().build_data(
            data_to_build=data_to_build,
            **kwargs,

        )
        data_to_build["SHOW_ORDER_ID"] = False

        self._build_footer(data_to_build=data_to_build)

        return data_to_build



import logging
import sqlite3
from sqlite3 import Connection
from typing import (
    Dict,
    Optional,
)

from boss.interactor.authenticate import AuthenticateInteractor
from boss.interactor.authenticate_nfc import AuthenticateNfcInteractor
from boss.interactor.get_info import GetInfoInteractor
from boss.interactor.get_users_list import GetUsersListInteractor
from boss.interactor.insert_or_update_users import InsertOrUpdateUsersInteractor
from boss.interactor.insert_or_update_users.dto import InsertOrUpdateUsersRequest
from boss.model.service import AuthenticationService
from boss.processor import (
    AuthenticateNfcProcessor,
    AuthenticateProcessor,
    GetInfoProcessor,
    GetUsersListProcessor,
    InsertOrUpdateUsersProcessor,
)
from boss_api.shared import token
from messagehandler import (
    MessageHandler,
    MessageHandlerBuilder,
)
from messageprocessor import (
    DefaultMessageProcessorExecutorFactory,
    MessageProcessorMessageHandler,
)
from messageprocessorutil import LoggingProcessorCallback
from repositoryimpl.sqlite import (
    SqliteConnectionTransactionCallback,
    UserRepositoryImpl,
)
from sqlite_alembic_util import (
    AlembicDatabaseInfo,
    SqliteAlembicHandler,
)
from . import Configs

logger = logging.getLogger(__name__)


class BossMessageHandlerBuilder(MessageHandlerBuilder):

    def __init__(
        self,
        configs: Configs,
        authenticate_service: AuthenticationService,
    ) -> None:

        self.configs = configs
        self.conn_mapping: Dict[MessageHandler, Connection] = {}
        self.alembic_handler: Optional[SqliteAlembicHandler] = None

        self.authenticate_service: AuthenticationService = authenticate_service

    def build_singletons(
        self,
    ) -> None:

        self.alembic_handler = SqliteAlembicHandler(
            database_infos=AlembicDatabaseInfo(
                name="users",
                conn_strings=[
                    self.configs.database_path,
                ]
            )
        )
        self.alembic_handler.upgrade_to_head()

        self._insert_default_users()

    def build_message_handler(
        self,
        message: str,
        event: str,
    ) -> MessageHandler:

        conn = sqlite3.connect(
            database=self.configs.database_path,
            uri=True,
        )
        user_db_repository = UserRepositoryImpl(connection=conn)

        authenticate_interactor = AuthenticateInteractor(
            user_repository=user_db_repository,
            authenticate_service=self.authenticate_service,
            otp_interval=self.configs.otp_interval,
        )
        authenticate_processor = AuthenticateProcessor(interactor=authenticate_interactor)
        authenticate_nfc_interactor = AuthenticateNfcInteractor(user_repository=user_db_repository)
        authenticate_nfc_processor = AuthenticateNfcProcessor(interactor=authenticate_nfc_interactor)
        get_info_interactor = GetInfoInteractor(user_repository=user_db_repository)
        get_info_processor = GetInfoProcessor(interactor=get_info_interactor)
        get_users_list_interactor = GetUsersListInteractor(user_repository=user_db_repository)
        get_users_list_processor = GetUsersListProcessor(interactor=get_users_list_interactor)
        insert_or_update_users_interactor = InsertOrUpdateUsersInteractor(user_repository=user_db_repository)
        insert_or_update_users_processor = InsertOrUpdateUsersProcessor(interactor=insert_or_update_users_interactor)

        message_handler = MessageProcessorMessageHandler(
            event_processors=None,
            message_processors={
                token.TK_USER_CONTROL_AUTHENTICATE: authenticate_processor,
                token.TK_USER_CONTROL_AUTHENTICATE_NFC: authenticate_nfc_processor,
                token.TK_USER_CONTROL_GET_INFO: get_info_processor,
                token.TK_USER_CONTROL_GET_USERS_LIST: get_users_list_processor,
                token.TK_USER_CONTROL_INSERT_OR_UPDATE_USERS: insert_or_update_users_processor,
            },
            executor_factory=DefaultMessageProcessorExecutorFactory(
                callbacks=[
                    LoggingProcessorCallback(logger=logging.getLogger("boss.LoggingProcessorCallback")),
                    SqliteConnectionTransactionCallback(conn=conn),
                ]
            ),
        )

        self.conn_mapping[message_handler] = conn

        return message_handler

    def destroy_message_handler(
        self,
        message_handler: MessageHandler,
    ) -> None:

        if message_handler in self.conn_mapping:
            conn = self.conn_mapping[message_handler]
            conn.close()
            del self.conn_mapping[message_handler]

    def destroy_singletons(
        self,
    ) -> None:

        return

    def _insert_default_users(
        self,
    ) -> None:

        conn = sqlite3.connect(
            database=self.configs.database_path,
            uri=True,
        )
        try:
            user_db_repository = UserRepositoryImpl(connection=conn)
            insert_or_update_users_interactor = InsertOrUpdateUsersInteractor(user_repository=user_db_repository)
            users_to_insert = InsertOrUpdateUsersRequest(users_to_insert_or_update=self.configs.default_users)
            insert_or_update_users_interactor.execute(request=users_to_insert)
            conn.commit()
        except Exception as ex:
            logger.exception("Error inserting default users: {}".format(str(ex)))
            conn.rollback()
        finally:
            conn.close()



from production.command import ProductionCommand


class UndoProductionCommand(ProductionCommand):
    def __init__(self, view):
        super(UndoProductionCommand, self).__init__(None, view)

    def get_hash_extra_value(self):
        return "UndoProductionCommand"



from typing import (
    Optional,
    Dict
)

from application.service import ChromeUpdaterService

from update_api.dto import ChromeUpdaterDto


class ChromeUpdaterInteractor(object):
    def __init__(
        self,
        chrome_updater_service  # type: ChromeUpdaterService
    ):
        # type: (...) -> None

        self.chrome_updater_service = chrome_updater_service

    def update_chrome(
        self,           # type: ChromeUpdaterInteractor
        validate,       # type: bool
        data=None,      # type: Optional[ChromeUpdaterDto]
    ):
        # type: (...) -> str

        return self.chrome_updater_service.update_chrome(
            validate_version=validate,
            restart_hv=True,
            data=data,
        )



import re

from production.model import TagEvent
from typing import List
from collections import OrderedDict

from _OrderChangerProductionBox import OrderChangerProductionBox

JIT_SEPARATOR = '[><$]'


class SequencerTagItemBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger=None):
        # type: (str, Optional[Union[List[ProductionBox]]], Logger) -> None # noqa

        super(SequencerTagItemBox, self).__init__(name, sons, logger)

    def change_order(self, order):
        for item in order.items:
            self._reprocess_tags(item)

        return order

    def _reprocess_tags(self, item):
        for son in item.items:
            self._reprocess_tags(son)

        if item.displayed_jit:
            self._verify_previous_jit_sequences(item)
            self._process_displayed_jit_tags(item)
            self._verify_next_jit_sequences(item)
            return

        self._create_generic_tags(item)

    def _process_displayed_jit_tags(self, item):
        for tag in item.tags:
            for item_jit in item.displayed_jit:
                if tag.endswith("-{}".format(item_jit.lower())):
                    new_tag = tag.replace("-{}".format(item_jit.lower()), "")
                    self._create_generic_tag_info(item, new_tag, tag)

    def _verify_next_jit_sequences(self, item):
        for item_jit in item.displayed_jit:
            sequence_groups = item.jit_lines.split("<")
            jit_index = self._get_sequence_index(item_jit, sequence_groups) + 1

            while len(sequence_groups[jit_index:]) > 0:
                next_jit_sequences = sequence_groups[jit_index:]
                next_jit_sequences = re.split('[>$]', "+".join(next_jit_sequences))[0]
                next_jit_sequences = re.findall('\\w+', next_jit_sequences)
                tag_doing = self._check_any_tag_in_item(item, next_jit_sequences, ["doing", "done", "served"])
                if tag_doing is not None:
                    self._create_generic_tag_info(item, "doing", tag_doing)
                    self._create_generic_tag_info(item, "doing-{}".format(item_jit.lower()), tag_doing)

                tag_done = self._check_any_tag_in_item(item, next_jit_sequences, ["done", "served"])
                if tag_done is not None:
                    self._create_generic_tag_info(item, "done", tag_done)
                    self._create_generic_tag_info(item, "done-{}".format(item_jit.lower()), tag_done)

                tag_served = self._check_any_tag_in_item(item, next_jit_sequences, ["served"])
                if tag_served is not None:
                    self._create_generic_tag_info(item, "served", tag_served)
                    self._create_generic_tag_info(item, "served-{}".format(item_jit.lower()), tag_served)

                jit_index += 1

    def _verify_previous_jit_sequences(self, item):
        for item_jit in item.displayed_jit:
            done_previous_jit_sequence = self._get_done_previous_jit_sequence(item, item_jit)
            doing_previous_jit_sequence = self._get_doing_previous_jit_sequence(item, item_jit)

            if doing_previous_jit_sequence:
                tag_doing = self._check_any_tag_in_item(item, doing_previous_jit_sequence, ["doing", "done", "served"])
                if tag_doing is not None:
                    self._create_generic_tag_info(item, "doing", tag_doing)

            if done_previous_jit_sequence:
                block_command_separator = item.jit_lines.find('${}'.format(item_jit))
                if block_command_separator >= 0:
                    item.block_tags = True

                tag_done = self._check_all_tag_in_item(item, done_previous_jit_sequence, ["done", "served"])
                if tag_done is not None:
                    self._create_generic_tag_info(item, "done", tag_done)
                    item.block_tags = False

                tag_served = self._check_all_tag_in_item(item, done_previous_jit_sequence, ["served"])
                if tag_served is not None:
                    self._create_generic_tag_info(item, "served", tag_served)

        if item.is_combo():
            some_child_has_block_tags = any([x.block_tags for x in item.items])
            item.block_tags |= some_child_has_block_tags

    def _get_done_previous_jit_sequence(self, item, item_jit):
        if not item.jit_lines:
            return None

        sequence_groups = item.jit_lines.split("<")
        jit_index = self._get_sequence_index(item_jit, sequence_groups)
        if jit_index >= len(sequence_groups):
            return None

        jit_sequences = sequence_groups[jit_index]
        jit_sequences = re.split(JIT_SEPARATOR, jit_sequences)
        sub_jit_index = self._get_sequence_index(item_jit, jit_sequences)
        if sub_jit_index == 0:
            if jit_index == 0:
                return

            previous_jit_string = sequence_groups[jit_index - 1]
            return re.findall('\\w+', previous_jit_string)

        previous_jit_string = "+".join(jit_sequences[:sub_jit_index])
        return re.findall('\\w+', previous_jit_string)

    def _get_doing_previous_jit_sequence(self, item, item_jit):
        if not item.jit_lines:
            return None

        sequence_groups = re.split(JIT_SEPARATOR, item.jit_lines)
        jit_index = self._get_sequence_index(item_jit, sequence_groups)
        if jit_index >= len(sequence_groups):
            return None

        previous_jit_sequences = sequence_groups[:jit_index]
        if not previous_jit_sequences:
            return

        previous_jit_string = re.split(JIT_SEPARATOR, "+".join(previous_jit_sequences))[-1]
        return re.findall('\\w+', previous_jit_string)

    @staticmethod
    def _get_sequence_index(item_jit, sequence_groups):
        jit_index = 0
        for jit_elem in sequence_groups:
            if item_jit in re.findall('\\w+', jit_elem):
                return jit_index

            jit_index += 1

        return jit_index

    def _create_generic_tags(self, item):
        self._insert_dont_need_cook_necessary_tag(item)
        self._insert_generic_doing_tag(item)
        self._insert_full_done_served_tag(item)

    def _insert_generic_doing_tag(self, item):
        necessary_jit_lines = re.findall('\\w+', item.jit_lines)
        tag_doing = self._check_any_tag_in_item(item, necessary_jit_lines, ["doing", "done", "served"])
        if tag_doing is not None:
            self._create_generic_tag_info(item, "doing", tag_doing)

    def _insert_full_done_served_tag(self, item):
        sequence_groups = re.split(JIT_SEPARATOR, item.jit_lines)
        last_sequence_group = sequence_groups[-1]
        self._insert_item_tag(item, last_sequence_group, ["done", "served"], "done")
        self._insert_item_tag(item, last_sequence_group, ["served"], "served")

    def _insert_dont_need_cook_necessary_tag(self, item):
        if not item.has_tag("dont-need-cook"):
            return

        sequence_groups = re.split(JIT_SEPARATOR, item.jit_lines)
        last_sequence_group = sequence_groups[0]
        self._insert_item_tag(item, last_sequence_group, ["done", "served"], "done")
        self._insert_item_tag(item, last_sequence_group, ["served"], "served")

    def _insert_item_tag(self, item, sequence_group, necessary_tags, tag_to_insert):
        necessary_jit_lines = re.findall('\\w+', sequence_group)
        tag = self._check_all_tag_in_item(item, necessary_jit_lines, necessary_tags)
        if tag is not None:
            self._create_generic_tag_info(item, tag_to_insert, tag)

    def _create_generic_tag_info(self, item, tag_type, tag_name):
        new_tags = OrderedDict()
        for k, v in item.tags.iteritems():
            new_tags[k] = v
            if k == tag_name:
                new_tags[tag_type] = tag_type

        item.tags = new_tags
        tag_history_entry_list = [x for x in item.tag_history if x.tag == tag_name]
        if tag_history_entry_list:
            tag_history_entry_point = tag_history_entry_list[0]
            item.tag_history.append(TagEvent(
                tag=tag_type,
                action=tag_history_entry_point.action,
                date=tag_history_entry_point.date,
                hallow=True
            ))

        if item.is_product():
            for son in item.items:
                self._create_generic_tag_info(son, tag_type, tag_name)

    @staticmethod
    def _check_all_tag_in_item(item, necessary_jit_lines, tag_types):
        last_tag_name = None
        for jit in necessary_jit_lines:
            for tag_type in tag_types:
                tag_name = "{}-{}".format(tag_type, jit.lower())
                if tag_name in item.tags:
                    last_tag_name = tag_name
                    break
            else:
                return None

        return last_tag_name

    @staticmethod
    def _check_any_tag_in_item(item, necessary_jit_lines, tag_types):
        for jit in necessary_jit_lines:
            for tag_type in tag_types:
                tag_name = "{}-{}".format(tag_type, jit.lower())
                if tag_name in item.tags:
                    return tag_name

        return None



import logging

from model import Report
from model.exceptions import ErrorParsingReportData
from nreports.repository import I18N
from nreports_api.shared.report_dto import ReportDto
from .dto import ReportJsonParserRequest

logger = logging.getLogger(name=__name__)


class ReportJsonParserInteractor(object):

    def __init__(
        self,
        i18n_repository: I18N,
    ) -> None:

        self.i18n_repository = i18n_repository

    def execute(
        self,
        request: ReportJsonParserRequest,
    ) -> ReportDto:

        logger.info(msg=f"Report json to parse: {request.report_json}")

        try:
            report = Report(request.report_json)
        except (Exception,) as ex:
            raise ErrorParsingReportData(error_message=str(ex))

        formatted_report = report.format_data(
            built_data={},
            i18n_repository=self.i18n_repository.build_i18n_repository(pos_id=request.pos_id or 0),
            l10n_repository=self.i18n_repository.build_l10n_repository(pos_id=request.pos_id or 0),
        )

        return ReportDto(formatted_report)



from __future__ import absolute_import

import optparse
import sys
import re
import os
from .diff import htmldiff

description = """\
"""

parser = optparse.OptionParser(
    usage="%prog [OPTIONS] FILE1 FILE2\n"
    "%prog --annotate [OPTIONS] INFO1 FILE1 INFO2 FILE2 ...",
    description=description,
    )

parser.add_option(
    '-o', '--output',
    metavar="FILE",
    dest="output",
    default="-",
    help="File to write the difference to",
    )

parser.add_option(
    '-a', '--annotation',
    action="store_true",
    dest="annotation",
    help="Do an annotation")

def main(args=None):
    if args is None:
        args = sys.argv[1:]
    options, args = parser.parse_args(args)
    if options.annotation:
        return annotate(options, args)
    if len(args) != 2:
        print('Error: you must give two files')
        parser.print_help()
        sys.exit(1)
    file1, file2 = args
    input1 = read_file(file1)
    input2 = read_file(file2)
    body1 = split_body(input1)[1]
    pre, body2, post = split_body(input2)
    result = htmldiff(body1, body2)
    result = pre + result + post
    if options.output == '-':
        if not result.endswith('\n'):
            result += '\n'
        sys.stdout.write(result)
    else:
        with open(options.output, 'wb') as f:
            f.write(result)

def read_file(filename):
    if filename == '-':
        c = sys.stdin.read()
    elif not os.path.exists(filename):
        raise OSError(
            "Input file %s does not exist" % filename)
    else:
        with open(filename, 'rb') as f:
            c = f.read()
    return c

body_start_re = re.compile(
    r"<body.*?>", re.I|re.S)
body_end_re = re.compile(
    r"</body.*?>", re.I|re.S)
    
def split_body(html):
    pre = post = ''
    match = body_start_re.search(html)
    if match:
        pre = html[:match.end()]
        html = html[match.end():]
    match = body_end_re.search(html)
    if match:
        post = html[match.start():]
        html = html[:match.start()]
    return pre, html, post

def annotate(options, args):
    print("Not yet implemented")
    sys.exit(1)
    



import json
from logging import Logger
from threading import Lock

from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Event
from msgbus import MBEasyContext

from externals.repository import AlertsRepository
from externals.repository.api import AlertsAPIRepository
from services.util import get_store_id


updating_configured_alerts = Lock()


class UpdateService(object):

    def __init__(
        self,
        message_bus,            # type: MbContextMessageBus
        mb_context,             # type: MBEasyContext
        alerts_repository,      # type: AlertsRepository
        alerts_api_repository,  # type: AlertsAPIRepository
        logger,                 # type: Logger
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.mb_context = mb_context
        self.alerts_repository = alerts_repository
        self.alerts_api_repository = alerts_api_repository
        self.logger = logger

    def update_configured_alerts(
        self,
    ):
        # type: (...) -> None

        global updating_configured_alerts

        with updating_configured_alerts:
            configured_alerts = self.alerts_api_repository.retrieve_configured_alerts()
            configured_alerts = json.loads(configured_alerts)
            self.alerts_repository.update_configured_alerts(configured_alerts)

    def update_configured_alerts_ack(
        self,
    ):
        # type: (...) -> None

        global updating_configured_alerts

        with updating_configured_alerts:
            subject = "BohChangeNotificationAck"
            store_id = get_store_id()
            self.logger.info(
                "Sending event %r from store %r",
                subject,
                store_id,
            )
            event = Event(
                subject=subject,
                event_type="",
                data=json.dumps({
                    "store_code": store_id
                })
            )
            self.message_bus.publish_event(event=event)



from ctypes import Union
from logging import Logger

from mw_helper import ensure_list
from typing import (
    List,
    Optional,
)

from ._ItemFilterBox import ItemFilterBox
from ..repository import ProductRepository


class ItemCustomParamsFilterBox(ItemFilterBox):

    def __init__(
        self,                   # type: ItemCustomParamsFilterBox
        name,                   # type: str
        sons,                   # type: Union[str, List[str]]
        product_repository,     # type: ProductRepository
        custom_params,          # type: List[str]
        inherit_comments,       # type: bool
        logger=None,            # type: Optional[Logger]
    ):
        # type: (...) -> None

        allowed_part_codes = self._get_allowed_values(
            product_repository=product_repository,
            custom_params=custom_params,
        )

        super(ItemCustomParamsFilterBox, self).__init__(
            name=name,
            sons=sons,
            allowed_part_codes=allowed_part_codes,
            forbidden_part_codes=[],
            inherit_comments=inherit_comments,
            logger=logger
        )

    @staticmethod
    def _get_allowed_values(
        product_repository,     # type: ProductRepository
        custom_params,          # type: List[str]
    ):
        # type: (...) -> List[int]

        allowed_part_codes = []
        custom_params_list = ensure_list(custom_params)
        products_with_custom_params = product_repository.get_products_with_custom_params(ensure_list(custom_params))
        for custom_param in custom_params_list:
            if custom_param in products_with_custom_params:
                allowed_part_codes.extend(products_with_custom_params.get(custom_param).keys())

        return allowed_part_codes



from typing import Dict

from ._TableError import TableError


class TabNotInValidTabNames(TableError):

    def __init__(
        self,       # type: TabNotInValidTabNames
        tab_name,   # type: str
    ):
        # type: (...) -> None
        super(TabNotInValidTabNames, self).__init__("$TAB_NOT_REGISTERED|{}".format(tab_name))

        self.tab_name = tab_name

    def get_extra_data(
        self,   # type: TabNotInValidTabNames
    ):
        # type: (...) -> Dict[str, str]

        return {
            "tabName": self.tab_name,
        }



# testing/pickleable.py
# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors


"""Classes used in pickling tests, need to be at the module level for
unpickling.
"""

from __future__ import annotations

from .entities import ComparableEntity
from ..schema import Column
from ..types import String


class User(ComparableEntity):
    pass


class Order(ComparableEntity):
    pass


class Dingaling(ComparableEntity):
    pass


class EmailUser(User):
    pass


class Address(ComparableEntity):
    pass


# TODO: these are kind of arbitrary....
class Child1(ComparableEntity):
    pass


class Child2(ComparableEntity):
    pass


class Parent(ComparableEntity):
    pass


class Screen:
    def __init__(self, obj, parent=None):
        self.obj = obj
        self.parent = parent


class Mixin:
    email_address = Column(String)


class AddressWMixin(Mixin, ComparableEntity):
    pass


class Foo:
    def __init__(self, moredata, stuff="im stuff"):
        self.data = "im data"
        self.stuff = stuff
        self.moredata = moredata

    __hash__ = object.__hash__

    def __eq__(self, other):
        return (
            other.data == self.data
            and other.stuff == self.stuff
            and other.moredata == self.moredata
        )


class Bar:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    __hash__ = object.__hash__

    def __eq__(self, other):
        return (
            other.__class__ is self.__class__
            and other.x == self.x
            and other.y == self.y
        )

    def __str__(self):
        return "Bar(%d, %d)" % (self.x, self.y)


class OldSchool:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, other):
        return (
            other.__class__ is self.__class__
            and other.x == self.x
            and other.y == self.y
        )


class OldSchoolWithoutCompare:
    def __init__(self, x, y):
        self.x = x
        self.y = y


class BarWithoutCompare:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __str__(self):
        return "Bar(%d, %d)" % (self.x, self.y)


class NotComparable:
    def __init__(self, data):
        self.data = data

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return NotImplemented

    def __ne__(self, other):
        return NotImplemented


class BrokenComparable:
    def __init__(self, data):
        self.data = data

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        raise NotImplementedError



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import typing

from cryptography.hazmat.primitives.asymmetric import dh

MIN_MODULUS_SIZE: int

class DHPrivateKey: ...
class DHPublicKey: ...
class DHParameters: ...

class DHPrivateNumbers:
    def __init__(self, x: int, public_numbers: DHPublicNumbers) -> None: ...
    def private_key(self, backend: typing.Any = None) -> dh.DHPrivateKey: ...
    @property
    def x(self) -> int: ...
    @property
    def public_numbers(self) -> DHPublicNumbers: ...

class DHPublicNumbers:
    def __init__(
        self, y: int, parameter_numbers: DHParameterNumbers
    ) -> None: ...
    def public_key(self, backend: typing.Any = None) -> dh.DHPublicKey: ...
    @property
    def y(self) -> int: ...
    @property
    def parameter_numbers(self) -> DHParameterNumbers: ...

class DHParameterNumbers:
    def __init__(self, p: int, g: int, q: int | None = None) -> None: ...
    def parameters(self, backend: typing.Any = None) -> dh.DHParameters: ...
    @property
    def p(self) -> int: ...
    @property
    def g(self) -> int: ...
    @property
    def q(self) -> int | None: ...

def generate_parameters(
    generator: int, key_size: int, backend: typing.Any = None
) -> dh.DHParameters: ...
def from_pem_parameters(
    data: bytes, backend: typing.Any = None
) -> dh.DHParameters: ...
def from_der_parameters(
    data: bytes, backend: typing.Any = None
) -> dh.DHParameters: ...



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import typing

from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import pkcs7

def serialize_certificates(
    certs: list[x509.Certificate],
    encoding: serialization.Encoding,
) -> bytes: ...
def encrypt_and_serialize(
    builder: pkcs7.PKCS7EnvelopeBuilder,
    encoding: serialization.Encoding,
    options: typing.Iterable[pkcs7.PKCS7Options],
) -> bytes: ...
def sign_and_serialize(
    builder: pkcs7.PKCS7SignatureBuilder,
    encoding: serialization.Encoding,
    options: typing.Iterable[pkcs7.PKCS7Options],
) -> bytes: ...
def load_pem_pkcs7_certificates(
    data: bytes,
) -> list[x509.Certificate]: ...
def load_der_pkcs7_certificates(
    data: bytes,
) -> list[x509.Certificate]: ...



# orm/sync.py
# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: allow-untyped-defs, allow-untyped-calls


"""private module containing functions used for copying data
between instances based on join conditions.

"""

from __future__ import annotations

from . import exc
from . import util as orm_util
from .base import PassiveFlag


def populate(
    source,
    source_mapper,
    dest,
    dest_mapper,
    synchronize_pairs,
    uowcommit,
    flag_cascaded_pks,
):
    source_dict = source.dict
    dest_dict = dest.dict

    for l, r in synchronize_pairs:
        try:
            # inline of source_mapper._get_state_attr_by_column
            prop = source_mapper._columntoproperty[l]
            value = source.manager[prop.key].impl.get(
                source, source_dict, PassiveFlag.PASSIVE_OFF
            )
        except exc.UnmappedColumnError as err:
            _raise_col_to_prop(False, source_mapper, l, dest_mapper, r, err)

        try:
            # inline of dest_mapper._set_state_attr_by_column
            prop = dest_mapper._columntoproperty[r]
            dest.manager[prop.key].impl.set(dest, dest_dict, value, None)
        except exc.UnmappedColumnError as err:
            _raise_col_to_prop(True, source_mapper, l, dest_mapper, r, err)

        # technically the "r.primary_key" check isn't
        # needed here, but we check for this condition to limit
        # how often this logic is invoked for memory/performance
        # reasons, since we only need this info for a primary key
        # destination.
        if (
            flag_cascaded_pks
            and l.primary_key
            and r.primary_key
            and r.references(l)
        ):
            uowcommit.attributes[("pk_cascaded", dest, r)] = True


def bulk_populate_inherit_keys(source_dict, source_mapper, synchronize_pairs):
    # a simplified version of populate() used by bulk insert mode
    for l, r in synchronize_pairs:
        try:
            prop = source_mapper._columntoproperty[l]
            value = source_dict[prop.key]
        except exc.UnmappedColumnError as err:
            _raise_col_to_prop(False, source_mapper, l, source_mapper, r, err)

        try:
            prop = source_mapper._columntoproperty[r]
            source_dict[prop.key] = value
        except exc.UnmappedColumnError as err:
            _raise_col_to_prop(True, source_mapper, l, source_mapper, r, err)


def clear(dest, dest_mapper, synchronize_pairs):
    for l, r in synchronize_pairs:
        if (
            r.primary_key
            and dest_mapper._get_state_attr_by_column(dest, dest.dict, r)
            not in orm_util._none_set
        ):
            raise AssertionError(
                f"Dependency rule on column '{l}' "
                "tried to blank-out primary key "
                f"column '{r}' on instance '{orm_util.state_str(dest)}'"
            )
        try:
            dest_mapper._set_state_attr_by_column(dest, dest.dict, r, None)
        except exc.UnmappedColumnError as err:
            _raise_col_to_prop(True, None, l, dest_mapper, r, err)


def update(source, source_mapper, dest, old_prefix, synchronize_pairs):
    for l, r in synchronize_pairs:
        try:
            oldvalue = source_mapper._get_committed_attr_by_column(
                source.obj(), l
            )
            value = source_mapper._get_state_attr_by_column(
                source, source.dict, l, passive=PassiveFlag.PASSIVE_OFF
            )
        except exc.UnmappedColumnError as err:
            _raise_col_to_prop(False, source_mapper, l, None, r, err)
        dest[r.key] = value
        dest[old_prefix + r.key] = oldvalue


def populate_dict(source, source_mapper, dict_, synchronize_pairs):
    for l, r in synchronize_pairs:
        try:
            value = source_mapper._get_state_attr_by_column(
                source, source.dict, l, passive=PassiveFlag.PASSIVE_OFF
            )
        except exc.UnmappedColumnError as err:
            _raise_col_to_prop(False, source_mapper, l, None, r, err)

        dict_[r.key] = value


def source_modified(uowcommit, source, source_mapper, synchronize_pairs):
    """return true if the source object has changes from an old to a
    new value on the given synchronize pairs

    """
    for l, r in synchronize_pairs:
        try:
            prop = source_mapper._columntoproperty[l]
        except exc.UnmappedColumnError as err:
            _raise_col_to_prop(False, source_mapper, l, None, r, err)
        history = uowcommit.get_attribute_history(
            source, prop.key, PassiveFlag.PASSIVE_NO_INITIALIZE
        )
        if bool(history.deleted):
            return True
    else:
        return False


def _raise_col_to_prop(
    isdest, source_mapper, source_column, dest_mapper, dest_column, err
):
    if isdest:
        raise exc.UnmappedColumnError(
            "Can't execute sync rule for "
            "destination column '%s'; mapper '%s' does not map "
            "this column.  Try using an explicit `foreign_keys` "
            "collection which does not include this column (or use "
            "a viewonly=True relation)." % (dest_column, dest_mapper)
        ) from err
    else:
        raise exc.UnmappedColumnError(
            "Can't execute sync rule for "
            "source column '%s'; mapper '%s' does not map this "
            "column.  Try using an explicit `foreign_keys` "
            "collection which does not include destination column "
            "'%s' (or use a viewonly=True relation)."
            % (source_column, source_mapper, dest_column)
        ) from err



from datetime import datetime
from xml.etree import cElementTree as eTree
from typing import Optional

from sysactions import (
    StopAction,
    show_messagebox,
)


datetime_format = "%Y-%m-%dT%H:%M:%S.%f"


def _get_last_stored_state(
    order_picture,    # type: eTree.Element
):
    # type: (...) -> Optional[datetime]

    stored_states = order_picture.findall(".//StateHistory/State[@state='STORED']")
    if not stored_states:
        return None

    return datetime.strptime(
        stored_states[-1].get("timestampGMT"),
        datetime_format,
    )


def validate_can_change_order(
    pos_id,                 # type: str
    order_picture,          # type: eTree.Element
    line_to_replace,        # type: str
    context,                # type: str
    line_number,            # type: str
    sale_type=None,         # type: Optional[str]
):
    # type: (...) -> None

    sale_type = order_picture.get("saleTypeDescr") if not sale_type and order_picture else sale_type
    if sale_type != "DRIVE_THRU":
        return

    last_stored_datetime = _get_last_stored_state(order_picture=order_picture)
    if last_stored_datetime is None:
        return

    sale_line_last_updated = None
    for sale_line in order_picture.findall(".//SaleLine"):
        is_line_change = _check_if_is_trying_to_change_line(
            sale_line=sale_line,
            line_number=line_number,
            line_to_replace=line_to_replace,
            context=context,
        )
        adding_modifier = _is_adding_modifier(
            sale_line=sale_line,
            context=context,
            line_number=line_number,
        )
        if is_line_change or adding_modifier:
            sale_line_last_updated = datetime.strptime(
                sale_line.get("createdAt"),
                datetime_format,
            )
            break

    if sale_line_last_updated and sale_line_last_updated <= last_stored_datetime:
        show_messagebox(
            posid=pos_id,
            message="$CANNOT_CHANGE_DRIVE_OPENED_ORDER",
            icon="error",
        )
        raise StopAction()


def _is_adding_modifier(
    sale_line,      # type: eTree.Element,
    context,        # type: str
    line_number,    # type: str
):
    # type: (...) -> bool

    line_item_id = "{}.{}".format(
        sale_line.get("itemId"),
        sale_line.get("partCode"),
    )
    return line_item_id in context and line_number == sale_line.get("lineNumber")


def _check_if_is_trying_to_change_line(
    sale_line,          # type: eTree.Element
    line_number,        # type: str
    line_to_replace,    # type: str
    context,            # type: str
):
    # type: (...) -> bool

    abs_item_id = "{}.{}".format(
        sale_line.get("itemId"),
        sale_line.get("partCode"),
    )
    is_item_replace = line_to_replace == abs_item_id and line_number == sale_line.get("lineNumber")

    same_line_number = line_number == sale_line.get("lineNumber")
    is_option_or_can_add = sale_line.get("itemType") in ["OPTION", "CANADD", "INGREDIENT"]
    is_item_add = not line_to_replace and context == abs_item_id and is_option_or_can_add and same_line_number

    return is_item_replace or is_item_add



from helper import ExtendedEnum


# noinspection SpellCheckingInspection
class FiscalPaymentTypes(ExtendedEnum):
    DINHEIRO = ("01", "Dinheiro")
    CHEQUE = ("02", "Cheque")
    CARTAO_CREDITO = ("03", "Cartão de Crédito")
    CARTAO_DEBITO = ("04", "Cartão de Débito")
    CREDITO_LOJA = ("05", "Crédito Loja")
    VALE_ALIMENTACAO = ("10", "Vale Alimentação")
    VALE_REFEICAO = ("11", "Vale Refeição")
    VALE_PRESENTE = ("12", "Vale Presente")
    VALE_COMBUSTIVEL = ("13", "Vale Combustível")
    BOLETO_BANCARIO = ("15", "Boleto Bancário")
    DEPOSITO_BANCARIO = ("16", "Depósito Bancário")
    PIX = ("17", "PIX")
    STATIC_PIX = ("20", "PIX Estático")
    TRANSFERENCIA_BANCARIA = ("18", "Transferência Bancária")
    PROGRAMA_DE_FIDELIDADE = ("19", "Programa de Fidelidade")
    SEM_PAGAMENTO = ("90", "Sem pagamento")
    OUTROS = ("99", "Outros")



from abc import (
    ABC,
    abstractmethod,
)

from ._EventPublishDetail import EventPublishDetail


class NewEventListener(ABC):
    @abstractmethod
    def new_event(self, publish_detail: EventPublishDetail):
        raise NotImplementedError()



from ._CannotAddProductWithMandatoryCustomizationOnTotal import CannotAddProductWithMandatoryCustomizationOnTotal
from ._OrderException import OrderException
from ._ProductBarcodeNotFound import ProductBarcodeNotFound
from ._ProductInRupture import ProductInRupture
from ._UnavailableProductStock import UnavailableProductStock



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class AnyValue(google.protobuf.message.Message):
    """AnyValue is used to represent any type of attribute value. AnyValue may contain a
    primitive value such as a string or integer or it may contain an arbitrary nested
    object containing arrays, key-value lists and primitives.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    STRING_VALUE_FIELD_NUMBER: builtins.int
    BOOL_VALUE_FIELD_NUMBER: builtins.int
    INT_VALUE_FIELD_NUMBER: builtins.int
    DOUBLE_VALUE_FIELD_NUMBER: builtins.int
    ARRAY_VALUE_FIELD_NUMBER: builtins.int
    KVLIST_VALUE_FIELD_NUMBER: builtins.int
    BYTES_VALUE_FIELD_NUMBER: builtins.int
    string_value: typing.Text = ...
    bool_value: builtins.bool = ...
    int_value: builtins.int = ...
    double_value: builtins.float = ...
    @property
    def array_value(self) -> global___ArrayValue: ...
    @property
    def kvlist_value(self) -> global___KeyValueList: ...
    bytes_value: builtins.bytes = ...
    def __init__(self,
        *,
        string_value : typing.Text = ...,
        bool_value : builtins.bool = ...,
        int_value : builtins.int = ...,
        double_value : builtins.float = ...,
        array_value : typing.Optional[global___ArrayValue] = ...,
        kvlist_value : typing.Optional[global___KeyValueList] = ...,
        bytes_value : builtins.bytes = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["array_value",b"array_value","bool_value",b"bool_value","bytes_value",b"bytes_value","double_value",b"double_value","int_value",b"int_value","kvlist_value",b"kvlist_value","string_value",b"string_value","value",b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["array_value",b"array_value","bool_value",b"bool_value","bytes_value",b"bytes_value","double_value",b"double_value","int_value",b"int_value","kvlist_value",b"kvlist_value","string_value",b"string_value","value",b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["value",b"value"]) -> typing.Optional[typing_extensions.Literal["string_value","bool_value","int_value","double_value","array_value","kvlist_value","bytes_value"]]: ...
global___AnyValue = AnyValue

class ArrayValue(google.protobuf.message.Message):
    """ArrayValue is a list of AnyValue messages. We need ArrayValue as a message
    since oneof in AnyValue does not allow repeated fields.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    VALUES_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___AnyValue]:
        """Array of values. The array may be empty (contain 0 elements)."""
        pass
    def __init__(self,
        *,
        values : typing.Optional[typing.Iterable[global___AnyValue]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["values",b"values"]) -> None: ...
global___ArrayValue = ArrayValue

class KeyValueList(google.protobuf.message.Message):
    """KeyValueList is a list of KeyValue messages. We need KeyValueList as a message
    since `oneof` in AnyValue does not allow repeated fields. Everywhere else where we need
    a list of KeyValue messages (e.g. in Span) we use `repeated KeyValue` directly to
    avoid unnecessary extra wrapping (which slows down the protocol). The 2 approaches
    are semantically equivalent.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    VALUES_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyValue]:
        """A collection of key/value pairs of key-value pairs. The list may be empty (may
        contain 0 elements).
        The keys MUST be unique (it is not allowed to have more than one
        value with the same key).
        """
        pass
    def __init__(self,
        *,
        values : typing.Optional[typing.Iterable[global___KeyValue]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["values",b"values"]) -> None: ...
global___KeyValueList = KeyValueList

class KeyValue(google.protobuf.message.Message):
    """KeyValue is a key-value pair that is used to store Span attributes, Link
    attributes, etc.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    KEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    key: typing.Text = ...
    @property
    def value(self) -> global___AnyValue: ...
    def __init__(self,
        *,
        key : typing.Text = ...,
        value : typing.Optional[global___AnyValue] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...
global___KeyValue = KeyValue

class InstrumentationScope(google.protobuf.message.Message):
    """InstrumentationScope is a message representing the instrumentation scope information
    such as the fully qualified name and version.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    ATTRIBUTES_FIELD_NUMBER: builtins.int
    DROPPED_ATTRIBUTES_COUNT_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    """An empty instrumentation scope name means the name is unknown."""

    version: typing.Text = ...
    @property
    def attributes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyValue]:
        """Additional attributes that describe the scope. [Optional].
        Attribute keys MUST be unique (it is not allowed to have more than one
        attribute with the same key).
        """
        pass
    dropped_attributes_count: builtins.int = ...
    def __init__(self,
        *,
        name : typing.Text = ...,
        version : typing.Text = ...,
        attributes : typing.Optional[typing.Iterable[global___KeyValue]] = ...,
        dropped_attributes_count : builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attributes",b"attributes","dropped_attributes_count",b"dropped_attributes_count","name",b"name","version",b"version"]) -> None: ...
global___InstrumentationScope = InstrumentationScope



import json
from typing import (
    Optional,
    List,
)

from xml.etree import cElementTree as eTree
from api.order import get_order_service
from bustoken import TK_SLCTRL_OMGR_REVERTFRACTIONLINE
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)
from posot import (
    OrderTaker,
    OrderTakerException,
)

from custom_sysactions import user_authorized_to_execute_action
from mw_helper import (
    show_messagebox,
    show_message_dialog,
)
from sysactions import (
    action,
    get_model,
    get_posot,
    has_current_order,
    send_message,
    StopAction,
)


title = "$INFORMATION"


@action
@user_authorized_to_execute_action
def do_merge_items(
    pos_id,            # type: str
    sale_line_json,    # type: json
):
    # type: (...) -> Optional[bool]
    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import create_table_modified_event

    sale_line_list = json.loads(sale_line_json)
    sale_line_list = filter(lambda x: x.get("level") == 0, sale_line_list)

    model = get_model(posid=pos_id)
    pos_ot = get_posot(model=model)

    try:
        merge_line = _get_merge_line(
            pos_id=pos_id,
            pos_ot=pos_ot,
            sale_line_list=sale_line_list,
        )
        if merge_line is None:
            return False

        _merge_items(
            pos_id=pos_id,
            merge_line=merge_line,
            pos_ot=pos_ot,
        )
    except OrderTakerException as ex:
        show_messagebox(
            posid=pos_id,
            message="$ERROR_CODE_INFO|%d|%s" % (ex.getErrorCode(), ex.getErrorDescr()),
        )
        raise StopAction()
    finally:
        pos_ot.blkopnotify = False

        model = get_model(pos_id)
        if has_current_order(model):
            order_service = get_order_service(pos_id=pos_id)
            order_service.store_order()

    table_id = tableService().get_current_table_id(pos_id)
    table_picture = tableService().get_table_picture(pos_id, table_id)

    create_table_modified_event(pos_id, table_picture)
    return True


def _get_merge_line(
    pos_id,            # type: str
    pos_ot,            # type: OrderTaker
    sale_line_list,    # type: List[dict]
):
    # type: (...) -> Optional[dict]

    invalid_line = _have_invalid_line(
        pos_id=pos_id,
        sale_line_list=sale_line_list,
    )
    if invalid_line:
        return None

    all_parts_is_selected = _check_if_all_parts_is_selected(
        pos_id=pos_id,
        pos_ot=pos_ot,
        sale_line_list=sale_line_list,
    )
    if not all_parts_is_selected:
        return None

    return sale_line_list[0]


def _check_if_all_parts_is_selected(
    pos_id,           # type: str
    pos_ot,           # type: OrderTaker
    sale_line_list,   # type: List[dict]
):
    # type: (...) -> bool

    order_picture = eTree.XML(pos_ot.orderPicture(pos_id, sale_line_list[0]["orderId"]))
    fraction_from = sale_line_list[0]["fractionFrom"]
    sale_lines = order_picture.findall(".//SaleLine")
    count_lines = 0
    for sale_line in sale_lines:
        custom_properties = json.loads(sale_line.get("customProperties"))
        if custom_properties.get("OMGR_FRACTION_FROM_LINE", 0) == fraction_from:
            count_lines += 1

    if len(sale_line_list) != count_lines:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$SELECT_ALL_PARTS",
        )
        return False

    return True


def _have_invalid_line(
    pos_id,            # type: str
    sale_line_list,    # type: List[dict]
):
    # type: (...) -> bool

    order_ids = []
    fractions_from = []
    for sale_line in sale_line_list:
        order_ids.append(int(sale_line["orderId"]))
        fractions_from.append(int(sale_line["fractionFrom"] or 0))

    unique_order_ids = list(set(order_ids))
    unique_fractions_from = list(set(fractions_from))

    if sum(fractions_from) == 0:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$NOT_FRACTIONATED_ITEM",
        )
        return True

    if len(unique_order_ids) > 1:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$CANNOT_MERGE",
        )
        return True

    if len(unique_fractions_from) > 1:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$SELECT_ALL_PARTS",
        )
        return True

    return False


def _merge_items(
    pos_id,        # type: str
    merge_line,    # type: dict
    pos_ot,        # type: OrderTaker
):
    # type: (...) -> None

    order_service = get_order_service(pos_id=pos_id)
    order_service.recall_order(order_id=merge_line["orderId"])
    block_notify = "1"
    data = "\0".join([
        block_notify,
        str(pos_id),
        str(merge_line['orderId']),
        str(merge_line["fractionFrom"]),
    ])
    msg = send_message(
        dest_name="ORDERMGR{}".format(pos_id),
        token=TK_SLCTRL_OMGR_REVERTFRACTIONLINE,
        format=FM_PARAM,
        data=data,
        timeout=60 * 1000000,
    )
    if msg.token != TK_SYS_ACK:
        show_message_dialog(
            pos_id=pos_id,
            title="$FATAL_ERROR_CALL_SUPPORT",
            message=msg.data,
        )
        raise StopAction()

    order_service.store_order()



# oib.py - functions for handling Croatian OIB numbers
# coding: utf-8
#
# Copyright (C) 2012, 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""OIB (Osobni identifikacijski broj, Croatian identification number).

The personal identification number is used to identify persons and legal
entities in Croatia. It has 11 digits (sometimes prefixed by HR), contains
no personal information and uses the ISO 7064 Mod 11, 10 checksum algorithm.

>>> validate('HR 33392005961')
'33392005961'
>>> validate('33392005962')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.iso7064 import mod_11_10
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -').upper().strip()
    if number.startswith('HR'):
        number = number[2:]
    return number


def validate(number):
    """Check if the number is a valid OIB number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    mod_11_10.validate(number)
    return number


def is_valid(number):
    """Check if the number is a valid OIB number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import gzip
import os
import xml.dom

from barcode import __release__

try:
    import Image
    import ImageDraw
    import ImageFont
except ImportError:
    try:
        from PIL import Image, ImageDraw, ImageFont  # lint:ok
    except ImportError:
        import logging
        log = logging.getLogger('pyBarcode')
        log.info('PIL not found. Image output disabled')
        Image = ImageDraw = ImageFont = None  # lint:ok


def mm2px(mm, dpi=300):
    return (mm * dpi) / 25.4


def pt2mm(pt):
    return pt * 0.352777778


def _set_attributes(element, **attributes):
    for key, value in attributes.items():
        element.setAttribute(key, value)


def create_svg_object():
    imp = xml.dom.getDOMImplementation()
    doctype = imp.createDocumentType(
        'svg',
        '-//W3C//DTD SVG 1.1//EN',
        'http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd'
    )
    document = imp.createDocument(None, 'svg', doctype)
    _set_attributes(document.documentElement, version='1.1',
                    xmlns='http://www.w3.org/2000/svg')
    return document


SIZE = '{0:.3f}mm'
COMMENT = 'Autogenerated with pyBarcode {0}'.format(__release__)
PATH = os.path.dirname(os.path.abspath(__file__))
FONT = os.path.join(PATH, 'DejaVuSansMono.ttf')


class BaseWriter(object):
    """Baseclass for all writers.

    Initializes the basic writer options. Childclasses can add more
    attributes and can set them directly or using
    `self.set_options(option=value)`.

    :parameters:
        initialize : Function
            Callback for initializing the inheriting writer.
            Is called: `callback_initialize(raw_code)`
        paint_module : Function
            Callback for painting one barcode module.
            Is called: `callback_paint_module(xpos, ypos, width, color)`
        paint_text : Function
            Callback for painting the text under the barcode.
            Is called: `callback_paint_text(xpos, ypos)` using `self.text`
            as text.
        finish : Function
            Callback for doing something with the completely rendered
            output.
            Is called: `return callback_finish()` and must return the
            rendered output.
    """

    def __init__(self, initialize=None, paint_module=None, paint_text=None,
                 finish=None):
        self._callbacks = dict(
            initialize=initialize, paint_module=paint_module,
            paint_text=paint_text, finish=finish
        )
        self.module_width = 10
        self.module_height = 10
        self.font_size = 10
        self.quiet_zone = 6.5
        self.background = 'white'
        self.foreground = 'black'
        self.text = ''
        self.human = ''  # human readable text
        self.text_distance = 5
        self.center_text = True

    def calculate_size(self, modules_per_line, number_of_lines, dpi=300):
        """Calculates the size of the barcode in pixel.

        :parameters:
            modules_per_line : Integer
                Number of modules in one line.
            number_of_lines : Integer
                Number of lines of the barcode.
            dpi : Integer
                DPI to calculate.

        :returns: Width and height of the barcode in pixel.
        :rtype: Tuple
        """
        width = 2 * self.quiet_zone + modules_per_line * self.module_width
        height = 2.0 + self.module_height * number_of_lines
        if self.font_size and self.text:
            height += pt2mm(self.font_size) / 2 + self.text_distance
        return int(mm2px(width, dpi)), int(mm2px(height, dpi))

    def save(self, filename, output):
        """Saves the rendered output to `filename`.

        :parameters:
            filename : String
                Filename without extension.
            output : String
                The rendered output.

        :returns: The full filename with extension.
        :rtype: String
        """
        raise NotImplementedError

    def register_callback(self, action, callback):
        """Register one of the three callbacks if not given at instance
        creation.

        :parameters:
            action : String
                One of 'initialize', 'paint_module', 'paint_text', 'finish'.
            callback : Function
                The callback function for the given action.
        """
        self._callbacks[action] = callback

    def set_options(self, options):
        """Sets the given options as instance attributes (only
        if they are known).

        :parameters:
            options : Dict
                All known instance attributes and more if the childclass
                has defined them before this call.

        :rtype: None
        """
        for key, val in options.items():
            key = key.lstrip('_')
            if hasattr(self, key):
                setattr(self, key, val)

    def render(self, code):
        """Renders the barcode to whatever the inheriting writer provides,
        using the registered callbacks.

        :parameters:
            code : List
                List of strings matching the writer spec
                (only contain 0 or 1).
        """
        if self._callbacks['initialize'] is not None:
            self._callbacks['initialize'](code)
        ypos = 1.0
        for cc, line in enumerate(code):
            """
            Pack line to list give better gfx result, otherwise in can
            result in aliasing gaps
            '11010111' -> [2, -1, 1, -1, 3]
            """
            line += ' '
            c = 1
            mlist = []
            for i in range(0, len(line) - 1):
                if line[i] == line[i+1]:
                    c += 1
                else:
                    if line[i] == "1":
                        mlist.append(c)
                    else:
                        mlist.append(-c)
                    c = 1
            # Left quiet zone is x startposition
            xpos = self.quiet_zone
            bxs = xpos  # x start of barcode
            for mod in mlist:
                if mod < 1:
                    color = self.background
                else:
                    color = self.foreground
                # remove painting for background colored tiles?
                self._callbacks['paint_module'](
                    xpos, ypos, self.module_width * abs(mod), color
                )
                xpos += self.module_width * abs(mod)
            bxe = xpos
            # Add right quiet zone to every line, except last line,
            # quiet zone already provided with background,
            # should it be removed complety?
            if (cc + 1) != len(code):
                self._callbacks['paint_module'](
                    xpos, ypos, self.quiet_zone, self.background
                )
            ypos += self.module_height
        if self.text and self._callbacks['paint_text'] is not None:
            ypos += self.text_distance
            if self.center_text:
                # better center position for text
                xpos = bxs + ((bxe - bxs) / 2.0)
            else:
                xpos = bxs
            self._callbacks['paint_text'](xpos, ypos)
        return self._callbacks['finish']()


class SVGWriter(BaseWriter):

    def __init__(self):
        BaseWriter.__init__(self, self._init, self._create_module,
                            self._create_text, self._finish)
        self.compress = False
        self.dpi = 25.4
        self._document = None
        self._root = None
        self._group = None

    def _init(self, code):
        width, height = self.calculate_size(len(code[0]), len(code), self.dpi)
        self._document = create_svg_object()
        self._root = self._document.documentElement
        attributes = dict(width=SIZE.format(width), height=SIZE.format(height))
        _set_attributes(self._root, **attributes)
        self._root.appendChild(self._document.createComment(COMMENT))
        # create group for easier handling in 3rd party software
        # like corel draw, inkscape, ...
        group = self._document.createElement('g')
        attributes = dict(id='barcode_group')
        _set_attributes(group, **attributes)
        self._group = self._root.appendChild(group)
        background = self._document.createElement('rect')
        attributes = dict(width='100%', height='100%',
                          style='fill:{0}'.format(self.background))
        _set_attributes(background, **attributes)
        self._group.appendChild(background)

    def _create_module(self, xpos, ypos, width, color):
        element = self._document.createElement('rect')
        attributes = dict(x=SIZE.format(xpos), y=SIZE.format(ypos),
                          width=SIZE.format(width),
                          height=SIZE.format(self.module_height),
                          style='fill:{0};'.format(color))
        _set_attributes(element, **attributes)
        self._group.appendChild(element)

    def _create_text(self, xpos, ypos):
        element = self._document.createElement('text')
        attributes = dict(x=SIZE.format(xpos), y=SIZE.format(ypos),
                          style='fill:{0};font-size:{1}pt;text-anchor:'
                                'middle;'.format(self.foreground,
                                                 self.font_size))
        _set_attributes(element, **attributes)
        # check option to override self.text with self.human (barcode as
        # human readable data, can be used to print own formats)
        if self.human != '':
            barcodetext = self.human
        else:
            barcodetext = self.text
        text_element = self._document.createTextNode(barcodetext)
        element.appendChild(text_element)
        self._group.appendChild(element)

    def _finish(self):
        if self.compress:
            return self._document.toxml(encoding='UTF-8')
        else:
            return self._document.toprettyxml(indent=4 * ' ', newl=os.linesep,
                                              encoding='UTF-8')

    def save(self, filename, output):
        if self.compress:
            _filename = '{0}.svgz'.format(filename)
            f = gzip.open(_filename, 'wb')
            f.write(output)
            f.close()
        else:
            _filename = '{0}.svg'.format(filename)
            with open(_filename, 'wb') as f:
                f.write(output)
        return _filename


if Image is None:
    ImageWriter = None
else:
    class ImageWriter(BaseWriter):

        def __init__(self):
            BaseWriter.__init__(self, self._init, self._paint_module,
                                self._paint_text, self._finish)
            self.format = 'PNG'
            self.dpi = 300
            self._image = None
            self._draw = None

        def _init(self, code):
            size = self.calculate_size(len(code[0]), len(code), self.dpi)
            self._image = Image.new('RGB', size, self.background)
            self._draw = ImageDraw.Draw(self._image)

        def _paint_module(self, xpos, ypos, width, color):
            size = [(mm2px(xpos, self.dpi), mm2px(ypos, self.dpi)),
                    (mm2px(xpos + width, self.dpi),
                     mm2px(ypos + self.module_height, self.dpi))]
            self._draw.rectangle(size, outline=color, fill=color)

        def _paint_text(self, xpos, ypos):
            font = ImageFont.truetype(FONT, self.font_size * 2)
            width, height = font.getsize(self.text)
            pos = (mm2px(xpos, self.dpi) - width // 2,
                   mm2px(ypos, self.dpi) - height // 4)
            self._draw.text(pos, self.text, font=font, fill=self.foreground)

        def _finish(self):
            return self._image

        def save(self, filename, output):
            filename = '{0}.{1}'.format(filename, self.format.lower())
            output.save(filename, self.format.upper())
            return filename



# mypy: no-warn-unused-ignores

from __future__ import annotations

from configparser import ConfigParser
import io
import os
import sys
import typing
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

if True:
    # zimports hack for too-long names
    from sqlalchemy.util import (  # noqa: F401
        inspect_getfullargspec as inspect_getfullargspec,
    )
    from sqlalchemy.util.compat import (  # noqa: F401
        inspect_formatargspec as inspect_formatargspec,
    )

is_posix = os.name == "posix"

py311 = sys.version_info >= (3, 11)
py310 = sys.version_info >= (3, 10)
py39 = sys.version_info >= (3, 9)


# produce a wrapper that allows encoded text to stream
# into a given buffer, but doesn't close it.
# not sure of a more idiomatic approach to this.
class EncodedIO(io.TextIOWrapper):
    def close(self) -> None:
        pass


if py39:
    from importlib import resources as _resources

    importlib_resources = _resources
    from importlib import metadata as _metadata

    importlib_metadata = _metadata
    from importlib.metadata import EntryPoint as EntryPoint
else:
    import importlib_resources  # type:ignore # noqa
    import importlib_metadata  # type:ignore # noqa
    from importlib_metadata import EntryPoint  # type:ignore # noqa


def importlib_metadata_get(group: str) -> Sequence[EntryPoint]:
    ep = importlib_metadata.entry_points()
    if hasattr(ep, "select"):
        return ep.select(group=group)
    else:
        return ep.get(group, ())  # type: ignore


def formatannotation_fwdref(
    annotation: Any, base_module: Optional[Any] = None
) -> str:
    """vendored from python 3.7"""
    # copied over _formatannotation from sqlalchemy 2.0

    if isinstance(annotation, str):
        return annotation

    if getattr(annotation, "__module__", None) == "typing":
        return repr(annotation).replace("typing.", "").replace("~", "")
    if isinstance(annotation, type):
        if annotation.__module__ in ("builtins", base_module):
            return repr(annotation.__qualname__)
        return annotation.__module__ + "." + annotation.__qualname__
    elif isinstance(annotation, typing.TypeVar):
        return repr(annotation).replace("~", "")
    return repr(annotation).replace("~", "")


def read_config_parser(
    file_config: ConfigParser,
    file_argument: Sequence[Union[str, os.PathLike[str]]],
) -> List[str]:
    if py310:
        return file_config.read(file_argument, encoding="locale")
    else:
        return file_config.read(file_argument)



from ._ServiceChild import ServiceChild
from ._ServiceTender import ServiceTender
from ._SimpleTable import SimpleTable
from ._Table import Table
from ._TableMode import TableMode
from ._TableServiceDto import TableServiceDto
from ._TableStatus import TableStatus
from .table_xml_parser import (
    parse_service_tenders,
    parse_table,
)



# pool/events.py
# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
from __future__ import annotations

import typing
from typing import Any
from typing import Optional
from typing import Type
from typing import Union

from .base import ConnectionPoolEntry
from .base import Pool
from .base import PoolProxiedConnection
from .base import PoolResetState
from .. import event
from .. import util

if typing.TYPE_CHECKING:
    from ..engine import Engine
    from ..engine.interfaces import DBAPIConnection


class PoolEvents(event.Events[Pool]):
    """Available events for :class:`_pool.Pool`.

    The methods here define the name of an event as well
    as the names of members that are passed to listener
    functions.

    e.g.::

        from sqlalchemy import event

        def my_on_checkout(dbapi_conn, connection_rec, connection_proxy):
            "handle an on checkout event"

        event.listen(Pool, 'checkout', my_on_checkout)

    In addition to accepting the :class:`_pool.Pool` class and
    :class:`_pool.Pool` instances, :class:`_events.PoolEvents` also accepts
    :class:`_engine.Engine` objects and the :class:`_engine.Engine` class as
    targets, which will be resolved to the ``.pool`` attribute of the
    given engine or the :class:`_pool.Pool` class::

        engine = create_engine("postgresql+psycopg2://scott:tiger@localhost/test")

        # will associate with engine.pool
        event.listen(engine, 'checkout', my_on_checkout)

    """  # noqa: E501

    _target_class_doc = "SomeEngineOrPool"
    _dispatch_target = Pool

    @util.preload_module("sqlalchemy.engine")
    @classmethod
    def _accept_with(
        cls,
        target: Union[Pool, Type[Pool], Engine, Type[Engine]],
        identifier: str,
    ) -> Optional[Union[Pool, Type[Pool]]]:
        if not typing.TYPE_CHECKING:
            Engine = util.preloaded.engine.Engine

        if isinstance(target, type):
            if issubclass(target, Engine):
                return Pool
            else:
                assert issubclass(target, Pool)
                return target
        elif isinstance(target, Engine):
            return target.pool
        elif isinstance(target, Pool):
            return target
        elif hasattr(target, "_no_async_engine_events"):
            target._no_async_engine_events()
        else:
            return None

    @classmethod
    def _listen(
        cls,
        event_key: event._EventKey[Pool],
        **kw: Any,
    ) -> None:
        target = event_key.dispatch_target

        kw.setdefault("asyncio", target._is_asyncio)

        event_key.base_listen(**kw)

    def connect(
        self,
        dbapi_connection: DBAPIConnection,
        connection_record: ConnectionPoolEntry,
    ) -> None:
        """Called at the moment a particular DBAPI connection is first
        created for a given :class:`_pool.Pool`.

        This event allows one to capture the point directly after which
        the DBAPI module-level ``.connect()`` method has been used in order
        to produce a new DBAPI connection.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        """

    def first_connect(
        self,
        dbapi_connection: DBAPIConnection,
        connection_record: ConnectionPoolEntry,
    ) -> None:
        """Called exactly once for the first time a DBAPI connection is
        checked out from a particular :class:`_pool.Pool`.

        The rationale for :meth:`_events.PoolEvents.first_connect`
        is to determine
        information about a particular series of database connections based
        on the settings used for all connections.  Since a particular
        :class:`_pool.Pool`
        refers to a single "creator" function (which in terms
        of a :class:`_engine.Engine`
        refers to the URL and connection options used),
        it is typically valid to make observations about a single connection
        that can be safely assumed to be valid about all subsequent
        connections, such as the database version, the server and client
        encoding settings, collation settings, and many others.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        """

    def checkout(
        self,
        dbapi_connection: DBAPIConnection,
        connection_record: ConnectionPoolEntry,
        connection_proxy: PoolProxiedConnection,
    ) -> None:
        """Called when a connection is retrieved from the Pool.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        :param connection_proxy: the :class:`.PoolProxiedConnection` object
          which will proxy the public interface of the DBAPI connection for the
          lifespan of the checkout.

        If you raise a :class:`~sqlalchemy.exc.DisconnectionError`, the current
        connection will be disposed and a fresh connection retrieved.
        Processing of all checkout listeners will abort and restart
        using the new connection.

        .. seealso:: :meth:`_events.ConnectionEvents.engine_connect`
           - a similar event
           which occurs upon creation of a new :class:`_engine.Connection`.

        """

    def checkin(
        self,
        dbapi_connection: Optional[DBAPIConnection],
        connection_record: ConnectionPoolEntry,
    ) -> None:
        """Called when a connection returns to the pool.

        Note that the connection may be closed, and may be None if the
        connection has been invalidated.  ``checkin`` will not be called
        for detached connections.  (They do not return to the pool.)

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        """

    @event._legacy_signature(
        "2.0",
        ["dbapi_connection", "connection_record"],
        lambda dbapi_connection, connection_record, reset_state: (
            dbapi_connection,
            connection_record,
        ),
    )
    def reset(
        self,
        dbapi_connection: DBAPIConnection,
        connection_record: ConnectionPoolEntry,
        reset_state: PoolResetState,
    ) -> None:
        """Called before the "reset" action occurs for a pooled connection.

        This event represents
        when the ``rollback()`` method is called on the DBAPI connection
        before it is returned to the pool or discarded.
        A custom "reset" strategy may be implemented using this event hook,
        which may also be combined with disabling the default "reset"
        behavior using the :paramref:`_pool.Pool.reset_on_return` parameter.

        The primary difference between the :meth:`_events.PoolEvents.reset` and
        :meth:`_events.PoolEvents.checkin` events are that
        :meth:`_events.PoolEvents.reset` is called not just for pooled
        connections that are being returned to the pool, but also for
        connections that were detached using the
        :meth:`_engine.Connection.detach` method as well as asyncio connections
        that are being discarded due to garbage collection taking place on
        connections before the connection was checked in.

        Note that the event **is not** invoked for connections that were
        invalidated using :meth:`_engine.Connection.invalidate`.    These
        events may be intercepted using the :meth:`.PoolEvents.soft_invalidate`
        and :meth:`.PoolEvents.invalidate` event hooks, and all "connection
        close" events may be intercepted using :meth:`.PoolEvents.close`.

        The :meth:`_events.PoolEvents.reset` event is usually followed by the
        :meth:`_events.PoolEvents.checkin` event, except in those
        cases where the connection is discarded immediately after reset.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        :param reset_state: :class:`.PoolResetState` instance which provides
         information about the circumstances under which the connection
         is being reset.

         .. versionadded:: 2.0

        .. seealso::

            :ref:`pool_reset_on_return`

            :meth:`_events.ConnectionEvents.rollback`

            :meth:`_events.ConnectionEvents.commit`

        """

    def invalidate(
        self,
        dbapi_connection: DBAPIConnection,
        connection_record: ConnectionPoolEntry,
        exception: Optional[BaseException],
    ) -> None:
        """Called when a DBAPI connection is to be "invalidated".

        This event is called any time the
        :meth:`.ConnectionPoolEntry.invalidate` method is invoked, either from
        API usage or via "auto-invalidation", without the ``soft`` flag.

        The event occurs before a final attempt to call ``.close()`` on the
        connection occurs.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        :param exception: the exception object corresponding to the reason
         for this invalidation, if any.  May be ``None``.

        .. seealso::

            :ref:`pool_connection_invalidation`

        """

    def soft_invalidate(
        self,
        dbapi_connection: DBAPIConnection,
        connection_record: ConnectionPoolEntry,
        exception: Optional[BaseException],
    ) -> None:
        """Called when a DBAPI connection is to be "soft invalidated".

        This event is called any time the
        :meth:`.ConnectionPoolEntry.invalidate`
        method is invoked with the ``soft`` flag.

        Soft invalidation refers to when the connection record that tracks
        this connection will force a reconnect after the current connection
        is checked in.   It does not actively close the dbapi_connection
        at the point at which it is called.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        :param exception: the exception object corresponding to the reason
         for this invalidation, if any.  May be ``None``.

        """

    def close(
        self,
        dbapi_connection: DBAPIConnection,
        connection_record: ConnectionPoolEntry,
    ) -> None:
        """Called when a DBAPI connection is closed.

        The event is emitted before the close occurs.

        The close of a connection can fail; typically this is because
        the connection is already closed.  If the close operation fails,
        the connection is discarded.

        The :meth:`.close` event corresponds to a connection that's still
        associated with the pool. To intercept close events for detached
        connections use :meth:`.close_detached`.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        """

    def detach(
        self,
        dbapi_connection: DBAPIConnection,
        connection_record: ConnectionPoolEntry,
    ) -> None:
        """Called when a DBAPI connection is "detached" from a pool.

        This event is emitted after the detach occurs.  The connection
        is no longer associated with the given connection record.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        :param connection_record: the :class:`.ConnectionPoolEntry` managing
         the DBAPI connection.

        """

    def close_detached(self, dbapi_connection: DBAPIConnection) -> None:
        """Called when a detached DBAPI connection is closed.

        The event is emitted before the close occurs.

        The close of a connection can fail; typically this is because
        the connection is already closed.  If the close operation fails,
        the connection is discarded.

        :param dbapi_connection: a DBAPI connection.
         The :attr:`.ConnectionPoolEntry.dbapi_connection` attribute.

        """



# veronumero.py - functions for handling Finnish individual tax numbers
# coding: utf-8
#
# Copyright (C) 2017 Holvi Payment Services Oy
# Copyright (C) 2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""
Veronumero (Finnish individual tax number).

The Veronumero an individual tax number that is assigned to workers in the
construction industry in Finland. The number is separate from the HETU and is
a 12 digit number without any embedded information such as birth dates.

More information:

* https://www.vero.fi/en/detailed-guidance/guidance/48791/individual_tax_numbers__instructions_fo/
* https://prosentti.vero.fi/Veronumerorekisteri/Tarkistus/VeronumeronTarkistus.aspx

>>> validate('123456789123')
'123456789123'
>>> validate('12345678912A')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('123456789')
Traceback (most recent call last):
    ...
InvalidLength: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the Veronumero to the minimal representation. This strips
    surrounding whitespace and removes separators."""
    return clean(number, ' ').strip()


def validate(number):
    """Check if the number is a valid tax number. This checks the length and
    formatting."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 12:
        raise InvalidLength()
    # there is no known check digit validation
    return number


def is_valid(number):
    """Check if the number is a valid tax number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import typing

from cryptography.hazmat.primitives.asymmetric import rsa

class RSAPrivateKey: ...
class RSAPublicKey: ...

class RSAPrivateNumbers:
    def __init__(
        self,
        p: int,
        q: int,
        d: int,
        dmp1: int,
        dmq1: int,
        iqmp: int,
        public_numbers: RSAPublicNumbers,
    ) -> None: ...
    @property
    def p(self) -> int: ...
    @property
    def q(self) -> int: ...
    @property
    def d(self) -> int: ...
    @property
    def dmp1(self) -> int: ...
    @property
    def dmq1(self) -> int: ...
    @property
    def iqmp(self) -> int: ...
    @property
    def public_numbers(self) -> RSAPublicNumbers: ...
    def private_key(
        self,
        backend: typing.Any = None,
        *,
        unsafe_skip_rsa_key_validation: bool = False,
    ) -> rsa.RSAPrivateKey: ...

class RSAPublicNumbers:
    def __init__(self, e: int, n: int) -> None: ...
    @property
    def n(self) -> int: ...
    @property
    def e(self) -> int: ...
    def public_key(self, backend: typing.Any = None) -> rsa.RSAPublicKey: ...

def generate_private_key(
    public_exponent: int,
    key_size: int,
) -> rsa.RSAPrivateKey: ...



# sedol.py - functions for handling SEDOL numbers
#
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""SEDOL number (Stock Exchange Daily Official List number).

The SEDOL number is a security identifier used in the United Kingdom and
Ireland assigned by the London Stock Exchange. A SEDOL is seven characters
in length consisting of six alphanumeric digits, followed by a check digit.

>>> validate('B15KXQ8')
'B15KXQ8'
>>> validate('B15KXQ7')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> to_isin('B15KXQ8')
'GB00B15KXQ89'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# the letters allowed in an SEDOL (vowels are never used)
_alphabet = '0123456789 BCD FGH JKLMN PQRST VWXYZ'


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip().upper()


def calc_check_digit(number):
    """Calculate the check digits for the number."""
    weights = (1, 3, 1, 7, 3, 9)
    s = sum(w * _alphabet.index(n) for w, n in zip(weights, number))
    return str((10 - s) % 10)


def validate(number):
    """Check if the number is valid. This checks the length and check
    digit."""
    number = compact(number)
    if not all(x in _alphabet for x in number):
        raise InvalidFormat()
    if len(number) != 7:
        raise InvalidLength()
    if isdigits(number[0]) and not isdigits(number):
        # new style SEDOLs are supposed to start with a letter, old-style
        # numbers should be fully numeric
        raise InvalidFormat()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is valid."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def to_isin(number):
    """Convert the number to an ISIN."""
    from stdnum import isin
    return isin.from_natid('GB', number)



from table_manager_api.dto.exception import TableError


class UnknownError(TableError):
    def __init__(
        self,       # type: UnknownError
        token,      # type: int
        data,       # type: bytes
    ):
        self.token = token
        try:
            self.message = data.decode()
        except UnicodeDecodeError:
            self.message = data.hex()

    def get_extra_data(self):
        return {
            "token": self.token,
            "data": self.message
        }

    def __str__(self):
        return "token: {}, data: {}".format(self.token, self.message)



# wkn.py - functions for handling Wertpapierkennnummer
# coding: utf-8
#
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Wertpapierkennnummer (German securities identification code).

The WKN, WPKN, WPK (Wertpapierkennnummer) is a German code to identify
securities. It is a 6-digit alphanumeric number without a check digit that no
longer has any structure. It is expected to be replaced by the ISIN.

>>> validate('A0MNRK')
'A0MNRK'
>>> validate('AOMNRK')  # no capital o allowed
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> to_isin('SKWM02')
'DE000SKWM021'
"""

from stdnum.exceptions import *
from stdnum.util import clean


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip().upper()


# O and I are not valid but are accounted for in the check digit calculation
_alphabet = '0123456789ABCDEFGH JKLMN PQRSTUVWXYZ'


def validate(number):
    """Check if the number provided is valid. This checks the length and
    check digit."""
    number = compact(number)
    if not all(x in _alphabet for x in number):
        raise InvalidFormat()
    if len(number) != 6:
        raise InvalidLength()
    return number


def is_valid(number):
    """Check if the number provided is valid. This checks the length and
    check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def to_isin(number):
    """Convert the number to an ISIN."""
    from stdnum import isin
    return isin.from_natid('DE', number)



from ._DiscountAppliedProcessor import DiscountAppliedProcessor
from ._GetConfiguredAlertsProcessor import GetConfiguredAlertsProcessor
from ._OrderCanceledProcessor import OrderCanceledProcessor
from ._POSAlertNotificationProcessor import POSAlertNotificationProcessor
from ._BOHChangeNotificationConfigProcessor import BOHChangeNotificationConfigProcessor
from ._BOHOrderPictureProcessor import BOHOrderPictureProcessor
from ._UpdateAlertsProcessor import UpdateAlertsProcessor



"""develop tests"""

import sys
from unittest import mock

import pytest

from setuptools import SetuptoolsDeprecationWarning
from setuptools.dist import Distribution


@pytest.mark.skipif(sys.platform == 'win32', reason='non-Windows only')
@pytest.mark.xfail(reason="bdist_rpm is long deprecated, should we remove it? #1988")
@mock.patch('distutils.command.bdist_rpm.bdist_rpm')
def test_bdist_rpm_warning(distutils_cmd, tmpdir_cwd):
    dist = Distribution(
        dict(
            script_name='setup.py',
            script_args=['bdist_rpm'],
            name='foo',
            py_modules=['hi'],
        )
    )
    dist.parse_command_line()
    with pytest.warns(SetuptoolsDeprecationWarning):
        dist.run_commands()

    distutils_cmd.run.assert_called_once()



"""

Revision ID: 8746dd15f3cb
Revises: 37c1f7921b35
Create Date: 2024-04-08 14:42:17.499202

"""
from typing import Sequence, Union

from alembic import op


# revision identifiers, used by Alembic.
revision: str = '8746dd15f3cb'
down_revision: Union[str, None] = '37c1f7921b35'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_index(
        index_name="PendingEvent_AckedAt",
        table_name="PendingEvent",
        columns=("AckedAt",),
    )

    op.create_index(
        index_name="PendingEvent_AbandonedAt",
        table_name="PendingEvent",
        columns=("AbandonedAt",),
    )


def downgrade() -> None:
    op.drop_index(
        index_name="PendingEvent_AckedAt",
        table_name="PendingEvent",
    )

    op.drop_index(
        index_name="PendingEvent_AbandonedAt",
        table_name="PendingEvent",
    )



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import typing

from cryptography.hazmat.primitives.asymmetric import ec

class ECPrivateKey: ...
class ECPublicKey: ...

class EllipticCurvePrivateNumbers:
    def __init__(
        self, private_value: int, public_numbers: EllipticCurvePublicNumbers
    ) -> None: ...
    def private_key(
        self, backend: typing.Any = None
    ) -> ec.EllipticCurvePrivateKey: ...
    @property
    def private_value(self) -> int: ...
    @property
    def public_numbers(self) -> EllipticCurvePublicNumbers: ...

class EllipticCurvePublicNumbers:
    def __init__(self, x: int, y: int, curve: ec.EllipticCurve) -> None: ...
    def public_key(
        self, backend: typing.Any = None
    ) -> ec.EllipticCurvePublicKey: ...
    @property
    def x(self) -> int: ...
    @property
    def y(self) -> int: ...
    @property
    def curve(self) -> ec.EllipticCurve: ...
    def __eq__(self, other: object) -> bool: ...

def curve_supported(curve: ec.EllipticCurve) -> bool: ...
def generate_private_key(
    curve: ec.EllipticCurve, backend: typing.Any = None
) -> ec.EllipticCurvePrivateKey: ...
def from_private_numbers(
    numbers: ec.EllipticCurvePrivateNumbers,
) -> ec.EllipticCurvePrivateKey: ...
def from_public_numbers(
    numbers: ec.EllipticCurvePublicNumbers,
) -> ec.EllipticCurvePublicKey: ...
def from_public_bytes(
    curve: ec.EllipticCurve, data: bytes
) -> ec.EllipticCurvePublicKey: ...
def derive_private_key(
    private_value: int, curve: ec.EllipticCurve
) -> ec.EllipticCurvePrivateKey: ...



import json

from typing import List, Optional, Dict

import pyscripts

from msgbus import FM_PARAM, TK_SYS_ACK
from bustoken import TK_PROD_GET_COURSES

from actions.models import Course

mb_context = pyscripts.mbcontext
cache_courses = None


def get_courses():
    # type: () -> Optional[List[Course]]

    global cache_courses

    if cache_courses is None:
        try:
            msg = mb_context.MB_EasySendMessage("ProductionSystem", TK_PROD_GET_COURSES, FM_PARAM, None, 10 * 1000000)
            if msg.token == TK_SYS_ACK:
                cache_courses = []
                for course in json.loads(msg.data):
                    cache_courses.append(_create_course(course))
            else:
                return []
        except Exception as _:
            return []

    return cache_courses


def _create_course(course):
    # type: (Dict) -> Course

    course_id = int(course.get("id"))
    course_name = course.get("name")
    course_description = course.get("description")
    course_abbreviation = course.get("abbreviation")
    course_delay = int(course.get("delay"))

    return Course(course_id, course_name, course_description, course_abbreviation, course_delay)



import unittest
import os
import importlib

from .compat.py39 import warnings_helper

import importlib_resources as resources

from . import util

# Since the functional API forwards to Traversable, we only test
# filesystem resources here -- not zip files, namespace packages etc.
# We do test for two kinds of Anchor, though.


class StringAnchorMixin:
    anchor01 = 'data01'
    anchor02 = 'data02'


class ModuleAnchorMixin:
    @property
    def anchor01(self):
        return importlib.import_module('data01')

    @property
    def anchor02(self):
        return importlib.import_module('data02')


class FunctionalAPIBase(util.DiskSetup):
    def setUp(self):
        super().setUp()
        self.load_fixture('data02')

    def _gen_resourcetxt_path_parts(self):
        """Yield various names of a text file in anchor02, each in a subTest"""
        for path_parts in (
            ('subdirectory', 'subsubdir', 'resource.txt'),
            ('subdirectory/subsubdir/resource.txt',),
            ('subdirectory/subsubdir', 'resource.txt'),
        ):
            with self.subTest(path_parts=path_parts):
                yield path_parts

    def assertEndsWith(self, string, suffix):
        """Assert that `string` ends with `suffix`.

        Used to ignore an architecture-specific UTF-16 byte-order mark."""
        self.assertEqual(string[-len(suffix) :], suffix)

    def test_read_text(self):
        self.assertEqual(
            resources.read_text(self.anchor01, 'utf-8.file'),
            'Hello, UTF-8 world!\n',
        )
        self.assertEqual(
            resources.read_text(
                self.anchor02,
                'subdirectory',
                'subsubdir',
                'resource.txt',
                encoding='utf-8',
            ),
            'a resource',
        )
        for path_parts in self._gen_resourcetxt_path_parts():
            self.assertEqual(
                resources.read_text(
                    self.anchor02,
                    *path_parts,
                    encoding='utf-8',
                ),
                'a resource',
            )
        # Use generic OSError, since e.g. attempting to read a directory can
        # fail with PermissionError rather than IsADirectoryError
        with self.assertRaises(OSError):
            resources.read_text(self.anchor01)
        with self.assertRaises(OSError):
            resources.read_text(self.anchor01, 'no-such-file')
        with self.assertRaises(UnicodeDecodeError):
            resources.read_text(self.anchor01, 'utf-16.file')
        self.assertEqual(
            resources.read_text(
                self.anchor01,
                'binary.file',
                encoding='latin1',
            ),
            '\x00\x01\x02\x03',
        )
        self.assertEndsWith(  # ignore the BOM
            resources.read_text(
                self.anchor01,
                'utf-16.file',
                errors='backslashreplace',
            ),
            'Hello, UTF-16 world!\n'.encode('utf-16-le').decode(
                errors='backslashreplace',
            ),
        )

    def test_read_binary(self):
        self.assertEqual(
            resources.read_binary(self.anchor01, 'utf-8.file'),
            b'Hello, UTF-8 world!\n',
        )
        for path_parts in self._gen_resourcetxt_path_parts():
            self.assertEqual(
                resources.read_binary(self.anchor02, *path_parts),
                b'a resource',
            )

    def test_open_text(self):
        with resources.open_text(self.anchor01, 'utf-8.file') as f:
            self.assertEqual(f.read(), 'Hello, UTF-8 world!\n')
        for path_parts in self._gen_resourcetxt_path_parts():
            with resources.open_text(
                self.anchor02,
                *path_parts,
                encoding='utf-8',
            ) as f:
                self.assertEqual(f.read(), 'a resource')
        # Use generic OSError, since e.g. attempting to read a directory can
        # fail with PermissionError rather than IsADirectoryError
        with self.assertRaises(OSError):
            resources.open_text(self.anchor01)
        with self.assertRaises(OSError):
            resources.open_text(self.anchor01, 'no-such-file')
        with resources.open_text(self.anchor01, 'utf-16.file') as f:
            with self.assertRaises(UnicodeDecodeError):
                f.read()
        with resources.open_text(
            self.anchor01,
            'binary.file',
            encoding='latin1',
        ) as f:
            self.assertEqual(f.read(), '\x00\x01\x02\x03')
        with resources.open_text(
            self.anchor01,
            'utf-16.file',
            errors='backslashreplace',
        ) as f:
            self.assertEndsWith(  # ignore the BOM
                f.read(),
                'Hello, UTF-16 world!\n'.encode('utf-16-le').decode(
                    errors='backslashreplace',
                ),
            )

    def test_open_binary(self):
        with resources.open_binary(self.anchor01, 'utf-8.file') as f:
            self.assertEqual(f.read(), b'Hello, UTF-8 world!\n')
        for path_parts in self._gen_resourcetxt_path_parts():
            with resources.open_binary(
                self.anchor02,
                *path_parts,
            ) as f:
                self.assertEqual(f.read(), b'a resource')

    def test_path(self):
        with resources.path(self.anchor01, 'utf-8.file') as path:
            with open(str(path), encoding='utf-8') as f:
                self.assertEqual(f.read(), 'Hello, UTF-8 world!\n')
        with resources.path(self.anchor01) as path:
            with open(os.path.join(path, 'utf-8.file'), encoding='utf-8') as f:
                self.assertEqual(f.read(), 'Hello, UTF-8 world!\n')

    def test_is_resource(self):
        is_resource = resources.is_resource
        self.assertTrue(is_resource(self.anchor01, 'utf-8.file'))
        self.assertFalse(is_resource(self.anchor01, 'no_such_file'))
        self.assertFalse(is_resource(self.anchor01))
        self.assertFalse(is_resource(self.anchor01, 'subdirectory'))
        for path_parts in self._gen_resourcetxt_path_parts():
            self.assertTrue(is_resource(self.anchor02, *path_parts))

    def test_contents(self):
        with warnings_helper.check_warnings((".*contents.*", DeprecationWarning)):
            c = resources.contents(self.anchor01)
        self.assertGreaterEqual(
            set(c),
            {'utf-8.file', 'utf-16.file', 'binary.file', 'subdirectory'},
        )
        with self.assertRaises(OSError), warnings_helper.check_warnings((
            ".*contents.*",
            DeprecationWarning,
        )):
            list(resources.contents(self.anchor01, 'utf-8.file'))

        for path_parts in self._gen_resourcetxt_path_parts():
            with self.assertRaises(OSError), warnings_helper.check_warnings((
                ".*contents.*",
                DeprecationWarning,
            )):
                list(resources.contents(self.anchor01, *path_parts))
        with warnings_helper.check_warnings((".*contents.*", DeprecationWarning)):
            c = resources.contents(self.anchor01, 'subdirectory')
        self.assertGreaterEqual(
            set(c),
            {'binary.file'},
        )

    @warnings_helper.ignore_warnings(category=DeprecationWarning)
    def test_common_errors(self):
        for func in (
            resources.read_text,
            resources.read_binary,
            resources.open_text,
            resources.open_binary,
            resources.path,
            resources.is_resource,
            resources.contents,
        ):
            with self.subTest(func=func):
                # Rejecting None anchor
                with self.assertRaises(TypeError):
                    func(None)
                # Rejecting invalid anchor type
                with self.assertRaises((TypeError, AttributeError)):
                    func(1234)
                # Unknown module
                with self.assertRaises(ModuleNotFoundError):
                    func('$missing module$')

    def test_text_errors(self):
        for func in (
            resources.read_text,
            resources.open_text,
        ):
            with self.subTest(func=func):
                # Multiple path arguments need explicit encoding argument.
                with self.assertRaises(TypeError):
                    func(
                        self.anchor02,
                        'subdirectory',
                        'subsubdir',
                        'resource.txt',
                    )


class FunctionalAPITest_StringAnchor(
    StringAnchorMixin,
    FunctionalAPIBase,
    unittest.TestCase,
):
    pass


class FunctionalAPITest_ModuleAnchor(
    ModuleAnchorMixin,
    FunctionalAPIBase,
    unittest.TestCase,
):
    pass
